searchNodes=[{"doc":"Machinist is a small library that allows you to implement finite state machines in a simple way. It provides a simple DSL to write combinations of transitions based on events.A good example is how we would implement the behaviour of a door. With machinist would be this way:defmodule Door do defstruct [state: :locked] use Machinist transitions do from :locked, to: :unlocked, event: &quot;unlock&quot; from :unlocked, to: :locked, event: &quot;lock&quot; from :unlocked, to: :opened, event: &quot;open&quot; from :opened, to: :closed, event: &quot;close&quot; from :closed, to: :opened, event: &quot;open&quot; from :closed, to: :locked, event: &quot;lock&quot; end endBy defining these rules with transitions and from macros, machinist generates and injects into the module Door transit/2 functions like this one:def transit(%Door{state: :locked} = struct, event: &quot;unlock&quot;) do {:ok, %Door{struct | state: :unlocked}} endSo that we can transit between states by relying on the state + event pattern matching.Let's see this in practice:By default, our Door is locked:iex&gt; door_locked = %Door{} %Door{state: :locked}So let's change its state to unlocked and opened:iex&gt; {:ok, door_unlocked} = Door.transit(door_locked, event: &quot;unlock&quot;) {:ok, %Door{state: :unlocked}} iex&gt; {:ok, door_opened} = Door.transit(door_unlocked, event: &quot;open&quot;) {:ok, %Door{state: :opened}}If we try to make a transition that does not follow the rules, we get an error:iex&gt; Door.transit(door_opened, event: &quot;lock&quot;) {:error, :not_allowed}Guard conditionsWe could also implement a state machine for an electronic door which should validate a passcode to unlock it. In this scenario, the machinist allows us to provide a function to evaluate a condition and return the new state.Check out the diagram below representing it:And to have this condition for the unlock event, use the event macro passing the guard option with a one-arity function:# .. transitions do event &quot;unlock&quot;, guard: &amp;check_passcode/1 do from :locked, to: :unlocked from :locked, to: :locked end end defp check_passcode(door) do if some_condition, do: :unlocked, else: :locked endSo when we call Door.transit(%Door{state: :locked}, event: &quot;unlock&quot;) the guard function check_passcode/1 will be called with the struct door as the first parameter and returns the new state to be set.Setting a different attribute name that holds the stateBy default, machinist expects the struct being updated to hold a state attribute, if you have state in a different attribute, pass the name as an atom, as follows:transitions attr: :door_state do # ... endAnd then machinist will set the state in that attribute.iex&gt; Door.transit(door, event: &quot;unlock&quot;) {:ok, %Door{door_state: :unlocked}}Implementing different versions of a state machineLet's suppose we want to build a selection process app that handles applications of candidates, and they may go through different versions of the process. For example:A Selection Process V1 with the following sequence of stages: [Registration] -&gt; [Code test] -&gt; [Enrollment]And a Selection Process V2 with these ones: [Registration] -&gt; [Interview] -&gt; [Enrollment]The difference here is in V1 candidates must take a Code Test and V2 an Interview.So, we could have a %Candidate{} struct that holds these attributes:defmodule SelectionProcess.Candidate do defstruct [:name, :state, test_score: 0] endAnd a SelectionProcess module that implements the state machine. Notice this time we don't want to implement the rules in the module that holds the state, in this case, it makes more sense for the SelectionProcess to keep the rules, also because we want more than one state machine version handling candidates as mentioned before. This is our V1 of the process:defmodule SelectionProcess.V1 do use Machinist alias SelectionProcess.Candidate @minimum_score 70 transitions Candidate do from :new, to: :registered, event: &quot;register&quot; from :registered, to: :started_test, event: &quot;start_test&quot; event &quot;send_test&quot;, guard: &amp;check_score/1 do from :started_test, to: :approved from :started_test, to: :reproved end from :approved, to: :enrolled, event: &quot;enroll&quot; end defp check_score(%Candidate{test_score: score}) do if score &gt;= @minimum_score, do: :approved, else: :reproved end endIn this code, we pass the Candidate module as a parameter to transitions to tell machinist that we expect V1.transit/2 functions with a %Candidate{} struct as first argument and not the %SelectionProcess.V1{} which would be by default.def transit(%Candidate{state: :new} = struct, event: &quot;register&quot;) do {:ok, %Candidate{struct | state: :registered}} endAlso notice we provided the function &amp;check_score/1 to the option to: instead of an atom, to decide the state based on the candidate test_score value.In version 2, we replaced the Code Test stage with the Interview which has different state transitions:defmodule SelectionProcess.V2 do use Machinist alias SelectionProcess.Candidate transitions Candidate do from :new, to: :registered, event: &quot;register&quot; from :registered, to: :interview_scheduled, event: &quot;schedule_interview&quot; from :interview_scheduled, to: :approved, event: &quot;approve_interview&quot; from :interview_scheduled, to: :repproved, event: &quot;reprove_interview&quot; from :approved, to: :enrolled, event: &quot;enroll&quot; end endNow let's see how we can test it:V1: A registered candidate wants to start his test.iex&gt; candidate1 = %Candidate{name: &quot;Ada&quot;, state: :registered} iex&gt; SelectionProcess.V1.transit(candidate1, event: &quot;start_test&quot;) %{:ok, %Candidate{state: :test_started}}V2: A registered candidate wants to schedule the interviewiex&gt; candidate2 = %Candidate{name: &quot;Jose&quot;, state: :registered} iex&gt; SelectionProcess.V2.transit(candidate1, event: &quot;schedule_interview&quot;) %{:ok, %Candidate{state: :interview_scheduled}}That's great because we also can implement many state machines for only one entity and test different scenarios, evaluate and collect data for deciding which one is better.machinist gives us this flexibility since it's just pure Elixir.Transiting from any state to anotherSometimes we need to define a from any state transition.Still, in the selection process example, candidates can abandon the process in a given state, and we want to be able to transit them to application_expired from any state. To do so, we just define a from with an underscore variable for the current state to be ignored.defmodule SelectionProcess.V2 do use Machinist alias SelectionProcess.Candidate transitions Candidate do # ... from _state, to: :application_expired, event: &quot;application_expired&quot; end endCode formatterElixir formatter (mix format) puts parenthesis around the macros.from(:some_state, to: :another, event: &quot;some_event&quot;) from :some_state do to(:another, event: &quot;some_event&quot;) endHowever, the machinist's.formatter.exs is configured to not use parenthesis. In order to follow the code style without parenthesis you can export the machinist config in your project.In your .formatter.exs file, just add the following:[ # ... import_deps: [:machinist] ]And you're good to go 🧙‍♀️.How does the DSL works?The use of transitions in combination with each from statement will be transformed in functions that will be injected into the module that is using machinist.This implementation:defmodule Door do defstruct state: :locked use Machinist transitions do from :locked, to: :unlocked, event: &quot;unlock&quot; from :unlocked, to: :locked, event: &quot;lock&quot; from :unlocked, to: :opened, event: &quot;open&quot; from :opened, to: :closed, event: &quot;close&quot; from :closed, to: :opened, event: &quot;open&quot; from :closed, to: :locked, event: &quot;lock&quot; end endIt is the same as:defmodule Door do defstruct state: :locked def transit(%__MODULE__{state: :locked} = struct, event: &quot;unlock&quot;) do {:ok, %__MODULE__{struct | state: :unlocked}} end def transit(%__MODULE__{state: :unlocked} = struct, event: &quot;lock&quot;) do {:ok, %__MODULE__{struct | state: :locked}} end def transit(%__MODULE__{state: :unlocked} = struct, event: &quot;open&quot;) do {:ok, %__MODULE__{struct | state: :opened}} end def transit(%__MODULE__{state: :opened} = struct, event: &quot;close&quot;) do {:ok, %__MODULE__{struct | state: :closed}} end def transit(%__MODULE__{state: :closed} = struct, event: &quot;open&quot;) do {:ok, %__MODULE__{struct | state: :opened}} end def transit(%__MODULE__{state: :closed} = struct, event: &quot;lock&quot;) do {:ok, %__MODULE__{struct | state: :locked}} end # a catchall function in case of unmatched clauses def transit(_, _), do: {:error, :not_allowed} endSo, as we can see, we can eliminate a lot of boilerplate with machinist making it easier to maintain and less prone to errors.","ref":"Machinist.html","title":"Machinist","type":"module"},{"doc":"Defines an event block grouping a same-event from -&gt; to transitionsevent &quot;form_submitted&quot; do from :form1, to: :form2 from :form2, to: :test end","ref":"Machinist.html#event/2","title":"Machinist.event/2","type":"macro"},{"doc":"Defines an event block grouping a same-event from -&gt; to transitions with a guard function that should evaluates a condition and returns a new state.event &quot;update_score&quot;&quot;, guard: &amp;check_score/1 do from :test, to: :approved from :test, to: :reproved end defp check_score(%{score: score}) do if score &gt;= 70, do: :approved, else: :reproved end","ref":"Machinist.html#event/3","title":"Machinist.event/3","type":"macro"},{"doc":"Defines a state transition with the given state, and the list of options [to: new_state, event: event]from 1, to: 2, event: &quot;next&quot;It's also possible to define a from any state transition to another specific one, by just passing an underscore variable in place of a real state valuefrom _state, to: :expired, event: &quot;enrollment_expired&quot;","ref":"Machinist.html#from/2","title":"Machinist.from/2","type":"macro"},{"doc":"Defines a block of transitions.By default transitions/1 expects the module using Machinist has a struct defined with a state attributetransitions do # ... end","ref":"Machinist.html#transitions/1","title":"Machinist.transitions/1","type":"macro"},{"doc":"Defines a block of transitions for a specific struct or defines a block of transitions just passing the attr option to define the attribute holding the stateExamplesA Candidate being handled by two different versions of a SelectionProcessdefmodule Candidate do defstruct state: :new end defmodule SelectionProcess.V1 do use Machinist transitions Candidate do from :new, to: :registered, event: &quot;register&quot; end end defmodule SelectionProcess.V2 do use Machinist transitions Candidate do from :new, to: :enrolled, event: &quot;enroll&quot; end endProviding the attr option to define the attribute holding the statedefmodule Candidate do defstruct candidate_state: :new use Machinist transitions attr: :candidate_state do from :new, to: :registered, event: &quot;register&quot; end end","ref":"Machinist.html#transitions/2","title":"Machinist.transitions/2","type":"macro"},{"doc":"Defines a block of transitions for a specific struct with attr option defining the attribute holding the statetransitions Candidate, attr: :candidate_state do # ... end","ref":"Machinist.html#transitions/3","title":"Machinist.transitions/3","type":"macro"},{"doc":"Machinist.Transition module behaviour","ref":"Machinist.Transition.html","title":"Machinist.Transition","type":"behaviour"},{"doc":"","ref":"Machinist.Transition.html#c:transit/2","title":"Machinist.Transition.transit/2","type":"callback"},{"doc":"","ref":"Machinist.NoLongerSupportedSyntaxError.html","title":"Machinist.NoLongerSupportedSyntaxError","type":"exception"},{"doc":"Machinist This small library allows you to implement finite-state machines with Elixir in a simple way. It provides a simple DSL to write combinations of transitions based on events.InstallationUsage","ref":"readme.html","title":"Machinist","type":"extras"},{"doc":"You can install machinist by adding it to your list of dependencies in mix.exs:def deps do [ {:machinist, &quot;~&gt; 1.0.0&quot;} ] end","ref":"readme.html#installation","title":"Machinist - Installation","type":"extras"},{"doc":"A good example is how we would implement the behaviour of a door. With machinist would be this way:defmodule Door do defstruct [state: :locked] use Machinist transitions do from :locked, to: :unlocked, event: &quot;unlock&quot; from :unlocked, to: :locked, event: &quot;lock&quot; from :unlocked, to: :opened, event: &quot;open&quot; from :opened, to: :closed, event: &quot;close&quot; from :closed, to: :opened, event: &quot;open&quot; from :closed, to: :locked, event: &quot;lock&quot; end endBy defining these rules with transitions and from macros, machinist generates and injects into the module Door transit/2 functions like this one:def transit(%Door{state: :locked} = struct, event: &quot;unlock&quot;) do {:ok, %Door{struct | state: :unlocked}} endSo that we can transit between states by relying on the state + event pattern matching.Let's see this in practice:By default, our Door is locked:iex&gt; door_locked = %Door{} %Door{state: :locked}So let's change its state to unlocked and opened:iex&gt; {:ok, door_unlocked} = Door.transit(door_locked, event: &quot;unlock&quot;) {:ok, %Door{state: :unlocked}} iex&gt; {:ok, door_opened} = Door.transit(door_unlocked, event: &quot;open&quot;) {:ok, %Door{state: :opened}}If we try to make a transition that does not follow the rules, we get an error:iex&gt; Door.transit(door_opened, event: &quot;lock&quot;) {:error, :not_allowed}","ref":"readme.html#usage","title":"Machinist - Usage","type":"extras"},{"doc":"We could also implement a state machine for an electronic door which should validate a passcode to unlock it. In this scenario, the machinist allows us to provide a function to evaluate a condition and return the new state.Check out the diagram below representing it:And to have this condition for the unlock event, use the event macro passing the guard option with a one-arity function:# .. transitions do event &quot;unlock&quot;, guard: &amp;check_passcode/1 do from :locked, to: :unlocked from :locked, to: :locked end end defp check_passcode(door) do if some_condition, do: :unlocked, else: :locked endSo when we call Door.transit(%Door{state: :locked}, event: &quot;unlock&quot;) the guard function check_passcode/1 will be called with the struct door as the first parameter and returns the new state to be set.Setting a different attribute name that holds the stateBy default, machinist expects the struct being updated to hold a state attribute, if you have state in a different attribute, pass the name as an atom, as follows:transitions attr: :door_state do # ... endAnd then machinist will set the state in that attribute.iex&gt; Door.transit(door, event: &quot;unlock&quot;) {:ok, %Door{door_state: :unlocked}}Implementing different versions of a state machineLet's suppose we want to build a selection process app that handles applications of candidates, and they may go through different versions of the process. For example:A Selection Process V1 with the following sequence of stages: [Registration] -&gt; [Code test] -&gt; [Enrollment]And a Selection Process V2 with these ones: [Registration] -&gt; [Interview] -&gt; [Enrollment]The difference here is in V1 candidates must take a Code Test and V2 an Interview.So, we could have a %Candidate{} struct that holds these attributes:defmodule SelectionProcess.Candidate do defstruct [:name, :state, test_score: 0] endAnd a SelectionProcess module that implements the state machine. Notice this time we don't want to implement the rules in the module that holds the state, in this case, it makes more sense for the SelectionProcess to keep the rules, also because we want more than one state machine version handling candidates as mentioned before. This is our V1 of the process:defmodule SelectionProcess.V1 do use Machinist alias SelectionProcess.Candidate @minimum_score 70 transitions Candidate do from :new, to: :registered, event: &quot;register&quot; from :registered, to: :started_test, event: &quot;start_test&quot; event &quot;send_test&quot;, guard: &amp;check_score/1 do from :started_test, to: :approved from :started_test, to: :reproved end from :approved, to: :enrolled, event: &quot;enroll&quot; end defp check_score(%Candidate{test_score: score}) do if score &gt;= @minimum_score, do: :approved, else: :reproved end endIn this code, we pass the Candidate module as a parameter to transitions to tell machinist that we expect V1.transit/2 functions with a %Candidate{} struct as first argument and not the %SelectionProcess.V1{} which would be by default.def transit(%Candidate{state: :new} = struct, event: &quot;register&quot;) do {:ok, %Candidate{struct | state: :registered}} endAlso notice we provided the function &amp;check_score/1 to the option to: instead of an atom, to decide the state based on the candidate test_score value.In version 2, we replaced the Code Test stage with the Interview which has different state transitions:defmodule SelectionProcess.V2 do use Machinist alias SelectionProcess.Candidate transitions Candidate do from :new, to: :registered, event: &quot;register&quot; from :registered, to: :interview_scheduled, event: &quot;schedule_interview&quot; from :interview_scheduled, to: :approved, event: &quot;approve_interview&quot; from :interview_scheduled, to: :repproved, event: &quot;reprove_interview&quot; from :approved, to: :enrolled, event: &quot;enroll&quot; end endNow let's see how we can test it:V1: A registered candidate wants to start his test.iex&gt; candidate1 = %Candidate{name: &quot;Ada&quot;, state: :registered} iex&gt; SelectionProcess.V1.transit(candidate1, event: &quot;start_test&quot;) %{:ok, %Candidate{state: :test_started}}V2: A registered candidate wants to schedule the interviewiex&gt; candidate2 = %Candidate{name: &quot;Jose&quot;, state: :registered} iex&gt; SelectionProcess.V2.transit(candidate1, event: &quot;schedule_interview&quot;) %{:ok, %Candidate{state: :interview_scheduled}}That's great because we also can implement many state machines for only one entity and test different scenarios, evaluate and collect data for deciding which one is better.machinist gives us this flexibility since it's just pure Elixir.Transiting from any state to anotherSometimes we need to define a from any state transition.Still, in the selection process example, candidates can abandon the process in a given state, and we want to be able to transit them to application_expired from any state. To do so, we just define a from with an underscore variable for the current state to be ignored.defmodule SelectionProcess.V2 do use Machinist alias SelectionProcess.Candidate transitions Candidate do # ... from _state, to: :application_expired, event: &quot;application_expired&quot; end endCode formatterElixir formatter (mix format) puts parenthesis around the macros.from(:some_state, to: :another, event: &quot;some_event&quot;) from :some_state do to(:another, event: &quot;some_event&quot;) endHowever, the machinist's.formatter.exs is configured to not use parenthesis. In order to follow the code style without parenthesis you can export the machinist config in your project.In your .formatter.exs file, just add the following:[ # ... import_deps: [:machinist] ]And you're good to go 🧙‍♀️.","ref":"readme.html#guard-conditions","title":"Machinist - Guard conditions","type":"extras"},{"doc":"The use of transitions in combination with each from statement will be transformed in functions that will be injected into the module that is using machinist.This implementation:defmodule Door do defstruct state: :locked use Machinist transitions do from :locked, to: :unlocked, event: &quot;unlock&quot; from :unlocked, to: :locked, event: &quot;lock&quot; from :unlocked, to: :opened, event: &quot;open&quot; from :opened, to: :closed, event: &quot;close&quot; from :closed, to: :opened, event: &quot;open&quot; from :closed, to: :locked, event: &quot;lock&quot; end endIt is the same as:defmodule Door do defstruct state: :locked def transit(%__MODULE__{state: :locked} = struct, event: &quot;unlock&quot;) do {:ok, %__MODULE__{struct | state: :unlocked}} end def transit(%__MODULE__{state: :unlocked} = struct, event: &quot;lock&quot;) do {:ok, %__MODULE__{struct | state: :locked}} end def transit(%__MODULE__{state: :unlocked} = struct, event: &quot;open&quot;) do {:ok, %__MODULE__{struct | state: :opened}} end def transit(%__MODULE__{state: :opened} = struct, event: &quot;close&quot;) do {:ok, %__MODULE__{struct | state: :closed}} end def transit(%__MODULE__{state: :closed} = struct, event: &quot;open&quot;) do {:ok, %__MODULE__{struct | state: :opened}} end def transit(%__MODULE__{state: :closed} = struct, event: &quot;lock&quot;) do {:ok, %__MODULE__{struct | state: :locked}} end # a catchall function in case of unmatched clauses def transit(_, _), do: {:error, :not_allowed} endSo, as we can see, we can eliminate a lot of boilerplate with machinist making it easier to maintain and less prone to errors.","ref":"readme.html#how-does-the-dsl-works","title":"Machinist - How does the DSL works?","type":"extras"},{"doc":"Feel free to contribute to this lib. If you have any suggestions or bug reports just open an issue or a PR.","ref":"readme.html#contributing","title":"Machinist - Contributing","type":"extras"},{"doc":"MIT License","ref":"readme.html#license","title":"Machinist - License","type":"extras"},{"doc":"If this project helps you reduce time to develop, you can give me a cup of coffee :)","ref":"readme.html#donate","title":"Machinist - Donate","type":"extras"}]